#include <assert.h>
#define OM_NDEBUG
#define NDEBUG
#include "mod2.h"
#include "structs.h" // Singular structs
#include "ring.h"
#include "numbers.h"
#include "polys.h"
#include "longrat.h"
#include "ideals.h"
#include "kstd1.h"
#undef NDEBUG

#include "polynomialring.h"
#include "polynomial.h"
#include "field_rationals.h"
#include "buchberger.h"
#include "division.h"
#include "printer.h"
#include "log.h"
#include <iostream>

const  char *singular_date=__DATE__" "__TIME__; 

static ring singularRing(PolynomialRing const &r)
{
  //  FieldRationalsImplementation *q=dynamic_cast<FieldRationalsImplementation>r.getField().implementingObject;

  ring ret=(ring)omAlloc0(sizeof(sip_sring));

  if(r.getField().isRationals())
    {
      ret->ch=0;
    }
  else
    {
      assert(0);
    }


  ret->N=r.getNumberOfVariables();

  ret->names=(char**) omAlloc(ret->N*sizeof(char*));
  for(int i=0;i<ret->N;i++)
    ret->names[i]=omStrDup(r.getVariableName(i).c_str());

  ret->order=(int*)omAlloc0(3*sizeof(int));
  ret->block0=(int*)omAlloc0(3*sizeof(int));
  ret->block1=(int*)omAlloc0(3*sizeof(int));

  //  ret->order[0]=ringorder_wp;//degree revlex
  ret->order[0]=ringorder_Wp;//degree lex
  ret->block0[0]=1;ret->block1[0]=ret->N;
  ret->order[1]=ringorder_C;

  ret->wvhdl=(int**)omAlloc0(3*sizeof(int*));
  ret->wvhdl[0]=(int*)omAlloc(ret->N*sizeof(int));
  for(int i=0;i<ret->N;i++)
    ret->wvhdl[0][i]=1;

  rComplete(ret);
  rChangeCurrRing(ret);

  return ret;
}

static void freeSingularRing(ring R)
{
  rKill(R);
}


static poly singularPolynomial(Polynomial const &p)
{
  int n=p.getRing().getNumberOfVariables();
  poly r=NULL;

  for(TermMap::const_iterator i=p.terms.begin();i!=p.terms.end();i++)
    {
      poly p=pInit();

      FieldElement c=i->second;

      mpq_t *C=fieldElementToGmp(c);
      number n2=nlRInit(0);
      mpz_init_set(&(n2)->z,mpq_numref(*C));
      mpz_init_set(&(n2)->n,mpq_denref(*C));
      n2->s=0;
      nNormalize(n2);
      pSetCoeff0(p,n2);

      for(int j=0;j<n;j++)
	pSetExp(p,j+1,i->first.exponent[j]);
      pSetComp(p,0);
      pSetm(p);
      r=pAdd(r,p);
    }

  return r;
}


static ideal singularPolynomialSet(PolynomialSet const &g)
{
  int m=g.size();


  ideal i=idInit(m,1);

  int J=0;
  for(PolynomialSet::const_iterator j=g.begin();j!=g.end();j++,J++)
    {
      i->m[J]=singularPolynomial(*j);
    }

  return i;
}

static FieldElement fromSingularCoefficient(PolynomialRing const &r, number c)
{
  FieldElement C(r.getField());
  
  //  if((nInt(c)==0) && ((SR_HDL(c) & SR_INT)==0))
  if(((SR_HDL(c) & SR_INT)==0))
    {
      mpq_t value;

      mpq_init(value);

      mpz_set(mpq_numref(value), &(c->z));
      if (c->s <3)
      mpz_set(mpq_denref(value), &(c->n));
      else
	 mpz_set_si(mpq_denref(value),1);
        
      C=fieldElementFromGmp(&value);
    }
  else
    {
      C=r.getField().zHomomorphism(nInt(c));
    }
  return C;
}


static Polynomial fromSingularPolynomial(PolynomialRing const &r, poly &p)
{
  int n=r.getNumberOfVariables();
  Polynomial ret(r);

  poly q=p;
  while(p)
    {
      IntegerVector v(n);
      for(int i=0;i<n;i++)
	v[i]=pGetExp(p,i+1);

      ret+=Term(fromSingularCoefficient(r,pGetCoeff(p)),Monomial(r,v));
      p=pNext(p);
    }
  if(q)
    {
      IntegerVector v(n);
      for(int i=0;i<n;i++)
	v[i]=pGetExp(q,i+1);

      ret.mark(Monomial(r,v));
    }

  return ret;
}

static PolynomialSet fromSingularIdeal(PolynomialRing const &r, ideal i)
{
  PolynomialSet ret(r);

  for(int j=0;j<IDELEMS(i);j++)
    if(i->m[j]!=NULL)
      ret.push_back(fromSingularPolynomial(r,i->m[j]));

  //  AsciiPrinter(Stderr).printPolynomialSet(ret);

  return ret;
}

void singularBuchberger(PolynomialSet *g, TermOrder const &termOrder)
{
  ring R=singularRing(g->getRing());

  ideal i=singularPolynomialSet(*g);

  //  ideal j=kStd(i,NULL,testHomog,NULL);
  test|=(Sy_bit(OPT_REDTAIL)|Sy_bit(OPT_INTSTRATEGY));

  ideal j=kStd(i,NULL,testHomog,NULL);
  //idShow(j);

  idDelete(&i);
  PolynomialSet ret=fromSingularIdeal(g->getRing(),j);
  idDelete(&j);

  
  freeSingularRing(R);

  *g=ret;
  //  return ret;
}

/**************************************************************************************************************************/

#include "groebnerengine.h"

static ring mySingularRingDegRevLex(PolynomialRing const &r, IntegerVector const &weight)
{
  //  FieldRationalsImplementation *q=dynamic_cast<FieldRationalsImplementation>r.getField().implementingObject;

  ring ret=(ring)omAlloc0(sizeof(sip_sring));

  if(r.getField().isRationals())
    {
      ret->ch=0;
    }
  else
    {
      assert(0);
    }


  ret->N=r.getNumberOfVariables();

  ret->names=(char**) omAlloc(ret->N*sizeof(char*));
  for(int i=0;i<ret->N;i++)
    ret->names[i]=omStrDup(r.getVariableName(i).c_str());

  ret->order=(int*)omAlloc0(3*sizeof(int));
  ret->block0=(int*)omAlloc0(3*sizeof(int));
  ret->block1=(int*)omAlloc0(3*sizeof(int));

  ret->order[0]=ringorder_wp;//degree revlex
  // ret->order[0]=ringorder_Wp;//degree lex
  ret->block0[0]=1;ret->block1[0]=ret->N;
  ret->order[1]=ringorder_C;

  ret->wvhdl=(int**)omAlloc0(3*sizeof(int*));
  ret->wvhdl[0]=(int*)omAlloc(ret->N*sizeof(int));
  for(int i=0;i<ret->N;i++)
    ret->wvhdl[0][i]=weight[i]+1000;//FIXMEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE

  rComplete(ret);
  rChangeCurrRing(ret);

  return ret;
}


class GroebnerEngineSingular : public GroebnerEngine
{
  virtual PolynomialSet groebnerBasis(bool &success, PolynomialSet const &idealGenerators, TermOrder const &termOrder, bool autoreduce)
  {
    PolynomialSet ret(idealGenerators.getRing());
    if(!ret.getRing().getField().isRationals())
      {
	success=false;
	return ret;
      }
    if(dynamic_cast<const WeightReverseLexicographicTermOrder*> (&termOrder))
      {
	const WeightReverseLexicographicTermOrder *T=dynamic_cast<const WeightReverseLexicographicTermOrder*> (&termOrder);

	ring R=mySingularRingDegRevLex(idealGenerators.getRing(),T->getWeight());
	ideal i=singularPolynomialSet(idealGenerators);
	test|=(Sy_bit(OPT_REDTAIL)|Sy_bit(OPT_INTSTRATEGY));

	log2 cerr<<"calling singular\n"; 
	ideal j=kStd(i,NULL,testHomog,NULL);
	log2 cerr<<"returning from singular\n"; 

	idDelete(&i);
	ret=fromSingularIdeal(ret.getRing(),j);
	idDelete(&j);
	freeSingularRing(R);

	ret.markAndScale(termOrder);
	minimize(&ret);
	if(autoreduce)
	  autoReduce_(&ret,termOrder);

	success=true;
	return ret;
      }
    success=false;
    return ret;
  }
  virtual PolynomialSet autoReduce(bool &success, PolynomialSet const &idealGenerators)
  {
    PolynomialSet ret(idealGenerators.getRing());
    if(!ret.getRing().getField().isRationals())
      {
	success=false;
	return ret;
      }
    
    IntegerVector weight=termorderWeight(idealGenerators);

    ring R=mySingularRingDegRevLex(idealGenerators.getRing(),weight);
    ideal i=singularPolynomialSet(idealGenerators);
    test|=(Sy_bit(OPT_REDTAIL)|Sy_bit(OPT_INTSTRATEGY));
    
    log2 cerr<<"calling singular\n"; 
    ideal j=kStd(i,NULL,testHomog,NULL);
	//    ideal j=kInterRed(i);
    log2 cerr<<"returning from singular\n"; 
    
    idDelete(&i);
    ret=fromSingularIdeal(ret.getRing(),j);
    idDelete(&j);
    freeSingularRing(R);
    
    ret.markAndScale(WeightTermOrder(weight));
    
    success=true;
    return ret;
  }
  virtual const char* name()
  {
    return "singular";
  }
};

static GroebnerEngineSingular groebnerEngineSingular;
